from django.db import models
from .querysets import APNSQuerySet, GCMQuerySet


class GCMDeviceManager(models.Manager):
    """
    A manager to be used for GCM Devices
    """

    def get_queryset(self):
        return GCMQuerySet(self.model)


class APNSDeviceManager(models.Manager):
    """
    A manager to be used for iOS devices
    """

    def get_queryset(self):
        APNSQuerySet(self.model)



from django.conf import settings


def gcm_deactivate(rids):
    pass

INSTAPUSH_SETTINGS = getattr(settings, 'INSTAPUSH_SETTINGS', {})

## GCM Settings
GCM_SETTINGS = INSTAPUSH_SETTINGS.get('GCM_SETTINGS', {})
GCM_SETTINGS.setdefault('API_URL', 'https://android.googleapis.com/gcm/send')
GCM_SETTINGS.setdefault('MAX_RECIPIENTS', 1000)
GCM_SETTINGS.setdefault('DEACTIVATE_UNREG_CALLBACK', gcm_deactivate)


## APNS Settings
APNS_SETTINGS = INSTAPUSH_SETTINGS.get('APNS_SETTINGS', {})
APNS_SETTINGS.setdefault('PORT', 2195)
APNS_SETTINGS.setdefault('FEEDBACK_PORT', 2196)
APNS_SETTINGS.setdefault('ERROR_TIMEOUT', None)
APNS_SETTINGS.setdefault('MAX_SIZE', 2048)

## Add host urls based on environment
if settings.DEBUG:
    APNS_SETTINGS.setdefault('HOST', 'gateway.sandbox.push.apple.com')
    APNS_SETTINGS.setdefault('FEEDBACK_HOST', 'feedback.sandbox.push.apple.com')
else:
    APNS_SETTINGS.setdefault('HOST', 'gateway.push.apple.com')
    APNS_SETTINGS.setdefault('FEEDBACK_HOST', 'feedback.push.apple.com')


INSTAPUSH_SETTINGS.setdefault('GCM_SETTINGS', GCM_SETTINGS)
INSTAPUSH_SETTINGS.setdefault('APNS_SETTINGS', APNS_SETTINGS)



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .fields import HexIntegerField
from .managers import  APNSDeviceManager, GCMDeviceManager


try:
    instapush_settings = settings.INSTAPUSH_SETTINGS
except AttributeError:
    raise ImproperlyConfigured("Please include instapush settings dictionary "\
            "in your django settings")


class BaseDevice(models.Model):
    """
    Represents a base device object. This class defines
    the generic fields to be used by all device types.
    All other device types should inherit from this.
    """

    name = models.CharField(_('name'), max_length=255, blank=True, null=True)
    active = models.BooleanField(_('active'), default=True)

    ## as a device can not only be related to a user
    ## but any other defined models. we let the push
    ## user decide which model should be the owner
    ## of a device object. For cases, where a device
    ## does not have to be related to any model this
    ## can be left empty and hence blank and null are
    ## set to True
    owner = models.ForeignKey(instapush_settings.get('DEVICE_OWNER_MODEL'),
            blank=True, null=True)

    created = models.DateTimeField(_('created'), auto_now_add=True)
    updated = models.DateTimeField(_('updated'), auto_now=True)

    class Meta:
        abstract = True


    def __unicode__(self):
        return self.name or self.device_id or self.registration_id


class GCMDevice(BaseDevice):
    """
    Represents an android device
    """

    device_id = HexIntegerField(_('Device ID'), blank=True, null=True,
            db_index=True)
    registration_id = models.TextField(_('Registration ID'))

    ## Set custom manager
    objects = GCMDeviceManager()

    class Meta:
        verbose_name = _('GCM Device')
        verbose_name_plural = _('GCM Devices')

    def send_message(self, message, **kwargs):
        """
        Sends a push notification to this device via GCM
        """

        from ..libs.gcm import gcm_send_message
        data = kwargs.pop("extra", {})
        if message is not None:
            data["message"] = message

        return gcm_send_message(registration_id=self.registration_id,
                data=data, **kwargs)


class APNSDevice(BaseDevice):
    """
    Represents an iOS device
    """

    device_id = models.UUIField(_('Device ID'), blank=True, null=True,
            db_index=True)
    registration_id = models.CharField(_('Registration ID'), max_length=64,
            unique=True)

    ## Set custom manager
    APNSDeviceManager()

    class Meta:
        verbose_name = _('APNS Device')
        verbose_name_plural = _('APNS Devices')

    def send_message(self, message, **kwargs):
        from ..libs.apns import apns_send_message
        return apns_send_message(registration_id=self.registration_id,
                alert=message, **kwargs)



"""
Defines the exception classes
"""


class PushError(Exception):
    def __init__(self, message=None):
        if message:
            self.message = message
        return super(GCMPushError, self).__init__(message)


class GCMPushError(PushError):
    pass


class APNSPushError(PushError):
    pass


class APNSDataOverflow(APNSPushError):
    pass

class APNSServerError(APNSPushError):
    def __init__(self, status, identifier):
        super(APNSServerError, self).__init__(status, identifier)
        self.status = status
        self.identifier = identifier



"""
Code by jleclanche @ https://github.com/jleclanche/django-push-notifications/blob/master/push_notifications/fields.py
"""

import re
import struct
from django import forms
from django.core.validators import RegexValidator
from django.db import models, connection
from django.utils import six
from django.utils.translation import ugettext_lazy as _


__all__ = ["HexadecimalField", "HexIntegerField"]


hex_re = re.compile(r"^(([0-9A-f])|(0x[0-9A-f]))+$")
postgres_engines = [
    "django.db.backends.postgresql_psycopg2",
    "django.contrib.gis.db.backends.postgis",
]


class HexadecimalField(forms.CharField):
    """
    A form field that accepts only hexadecimal numbers
    """
    def __init__(self, *args, **kwargs):
        self.default_validators = [RegexValidator(hex_re, _("Enter a valid hexadecimal number"), "invalid")]
        super(HexadecimalField, self).__init__(*args, **kwargs)

    def prepare_value(self, value):
        # converts bigint from db to hex before it is displayed in admin
        if value and not isinstance(value, six.string_types) \
            and connection.vendor in ("mysql", "sqlite"):
            value = hex(value).rstrip("L")
        return super(forms.CharField, self).prepare_value(value)


class HexIntegerField(six.with_metaclass(models.SubfieldBase, models.BigIntegerField)):
    """
    This field stores a hexadecimal *string* of up to 64 bits as an unsigned integer
    on *all* backends including postgres.

    Reasoning: Postgres only supports signed bigints. Since we don't care about
    signedness, we store it as signed, and cast it to unsigned when we deal with
    the actual value (with struct)

    On sqlite and mysql, native unsigned bigint types are used. In all cases, the
    value we deal with in python is always in hex.
    """
    def db_type(self, connection):
        engine = connection.settings_dict["ENGINE"]
        if "mysql" in engine:
            return "bigint unsigned"
        elif "sqlite" in engine:
            return "UNSIGNED BIG INT"
        else:
            return super(HexIntegerField, self).db_type(connection=connection)

    def get_prep_value(self, value):
        if value is None or value == "":
            return None
        if isinstance(value, six.string_types):
            value = int(value, 16)
        # on postgres only, interpret as signed
        if connection.settings_dict["ENGINE"] in postgres_engines:
            value = struct.unpack("q", struct.pack("Q", value))[0]
        return value

    def to_python(self, value):
        if isinstance(value, six.string_types):
            return value
        if value is None:
            return ""
        # on postgres only, re-interpret from signed to unsigned
        if connection.settings_dict["ENGINE"] in postgres_engines:
            value = hex(struct.unpack("Q", struct.pack("q", value))[0])
        return value

    def formfield(self, **kwargs):
        defaults = {"form_class": HexadecimalField}
        defaults.update(kwargs)
        # yes, that super call is right
        return super(models.IntegerField, self).formfield(**defaults)

    def run_validators(self, value):
        # make sure validation is performed on integer value not string value
        return super(models.BigIntegerField, self).run_validators(self.get_prep_value(value))



"""
Apple Push Notification Service
"""

import codecs
import json
import ssl
import struct
import socket
import time
from contextlib import closing
from binascii import unhexlify

from django.core.exceptions import ImproperlyConfigured
from ..exceptions import (
    APNSServerError,
    APNSDataOverflow,
)
from ..settings import INSTAPUSH_SETTINGS as settings


SETTINGS = settings.get('APNS_SETTINGS')


def _apns_create_socket(address_tuple):
    certfile = SETTINGS.get("APNS_CERTIFICATE")
    if not certfile:
        raise ImproperlyConfigured(
            'You need to set PUSH_NOTIFICATIONS_SETTINGS["APNS_CERTIFICATE"] to send messages through APNS.'
        )

    try:
        with open(certfile, "r") as f:
            f.read()
    except Exception as e:
        raise ImproperlyConfigured("The APNS certificate file at %r is not readable: %s" % (certfile, e))

    ca_certs = SETTINGS.get("APNS_CA_CERTIFICATES")

    sock = socket.socket()
    sock = ssl.wrap_socket(sock, ssl_version=ssl.PROTOCOL_TLSv1, certfile=certfile, ca_certs=ca_certs)
    sock.connect(address_tuple)

    return sock


def _apns_create_socket_to_push():
    return _apns_create_socket((SETTINGS["HOST"], SETTINGS["PORT"]))


def _apns_create_socket_to_feedback():
    return _apns_create_socket((SETTINGS["FEEDBACK_HOST"], SETTINGS["FEEDBACK_PORT"]))


def _apns_pack_frame(token_hex, payload, identifier, expiration, priority):
    token = unhexlify(token_hex)
    # |COMMAND|FRAME-LEN|{token}|{payload}|{id:4}|{expiration:4}|{priority:1}
    frame_len = 3 * 5 + len(token) + len(payload) + 4 + 4 + 1  # 5 items, each 3 bytes prefix, then each item length
    frame_fmt = "!BIBH%ssBH%ssBHIBHIBHB" % (len(token), len(payload))
    frame = struct.pack(
        frame_fmt,
        2, frame_len,
        1, len(token), token,
        2, len(payload), payload,
        3, 4, identifier,
        4, 4, expiration,
        5, 1, priority)

    return frame


def _apns_check_errors(sock):
    timeout = SETTINGS["ERROR_TIMEOUT"]
    if timeout is None:
        return  # assume everything went fine!
    saved_timeout = sock.gettimeout()
    try:
        sock.settimeout(timeout)
        data = sock.recv(6)
        if data:
            command, status, identifier = struct.unpack("!BBI", data)
            # apple protocol says command is always 8. See http://goo.gl/ENUjXg
            assert command == 8, "Command must be 8!"
            if status != 0:
                raise APNSServerError(status, identifier)
    except socket.timeout:  # py3, see http://bugs.python.org/issue10272
        pass
    except ssl.SSLError as e:  # py2
        if "timed out" not in e.message:
            raise
    finally:
        sock.settimeout(saved_timeout)


def _apns_send(token, alert, badge=None, sound="default", category=None, content_available=True,
    action_loc_key=None, loc_key=None, loc_args=[], extra={}, identifier=0,
    expiration=None, priority=10, socket=None):
    data = {}
    aps_data = {}

    custom_params = alert
    alert = custom_params.pop('message')

    if action_loc_key or loc_key or loc_args:
        alert = {"body": alert} if alert else {}
        if action_loc_key:
            alert["action-loc-key"] = action_loc_key
        if loc_key:
            alert["loc-key"] = loc_key
        if loc_args:
            alert["loc-args"] = loc_args

    if alert is not None:
        aps_data["alert"] = alert

    if badge is not None:
        aps_data["badge"] = badge

    if sound is not None:
        aps_data["sound"] = sound

    if category is not None:
        aps_data["category"] = category

    if content_available:
        aps_data["content-available"] = 1

    if custom_params.keys():
        aps_data['custom_params'] = {'data': custom_params}

    data["aps"] = aps_data
    data.update(extra)

    # convert to json, avoiding unnecessary whitespace with separators (keys sorted for tests)
    json_data = json.dumps(data, separators=(",", ":"), sort_keys=True).encode("utf-8")

    max_size = SETTINGS["MAX_SIZE"]
    if len(json_data) > max_size:
        raise APNSDataOverflow("Notification body cannot exceed %i bytes" % (max_size))

    # if expiration isn't specified use 1 month from now
    expiration_time = expiration if expiration is not None else int(time.time()) + 2592000

    frame = _apns_pack_frame(token, json_data, identifier, expiration_time, priority)

    if socket:
        socket.write(frame)
    else:
        with closing(_apns_create_socket_to_push()) as socket:
            socket.write(frame)
            _apns_check_errors(socket)


def _apns_read_and_unpack(socket, data_format):
    length = struct.calcsize(data_format)
    data = socket.recv(length)
    if data:
        return struct.unpack_from(data_format, data, 0)
    else:
        return None


def _apns_receive_feedback(socket):
    expired_token_list = []

    header_format = '!LH'
    has_data = True
    while has_data:
        try:
            header_data = _apns_read_and_unpack(socket, header_format)
            if header_data is not None:
                timestamp, token_length = header_data
                token_format = '%ss' % token_length
                device_token = _apns_read_and_unpack(socket, token_format)
                if device_token is not None:
                    expired_token_list.append((timestamp, device_token[0]))
            else:
                has_data = False
        except socket.timeout:
            pass
        except ssl.SSLError as e:
            if "timed out" not in e.message:
                raise

    return expired_token_list


def apns_send_message(registration_id, alert, **kwargs):
    """
    Sends an APNS notification to a single registration_id.
    This will send the notification as form data.
    If sending multiple notifications, it is more efficient to use
    apns_send_bulk_message()

    Note that if set alert should always be a string. If it is not set,
    it won't be included in the notification. You will need to pass None
    to this for silent notifications.
    """

    _apns_send(registration_id, alert, **kwargs)


def apns_send_bulk_message(registration_ids, alert, **kwargs):
    """
    Sends an APNS notification to one or more registration_ids.
    The registration_ids argument needs to be a list.

    Note that if set alert should always be a string. If it is not set,
    it won't be included in the notification. You will need to pass None
    to this for silent notifications.
    """
    with closing(_apns_create_socket_to_push()) as socket:
        for identifier, registration_id in enumerate(registration_ids):
            _apns_send(registration_id, alert, identifier=identifier, socket=socket, **kwargs)
        _apns_check_errors(socket)


def apns_fetch_inactive_ids():
    """
    Queries the APNS server for id's that are no longer active since
    the last fetch
    """
    with closing(_apns_create_socket_to_feedback()) as socket:
        inactive_ids = []

        for _, registration_id in _apns_receive_feedback(socket):
            inactive_ids.append(codecs.encode(registration_id, 'hex_codec'))

        return inactive_ids



import importlib


def get_model(module_location):
    """
    Returns the instance of the given module location.
    """

    if not isinstance(module_location, (str, unicode)):
        raise ValueError("The value provided should either be a string or "\
                "unicode instance. The value '%s' provided was %s "\
                "rather." % (module_location, type(module_location)))

    try:
        name_split = module_location.split(".")
        class_name = name_split.pop(-1)

        if not len(name_split):
            raise ValueError("The value should provide the module location "\
                    "joined by '.' e.g. for model named 'test' in "
                    "/app/module.py, The value should be 'app.module.test'")

        module_location = ".".join(name_split)
        module = importlib.import_module(module_location)
        cls = getattr(module, class_name)

        return cls

    except AttributeError:
        pass



# _*_ coding: utf-8 _*_ 
from distutils.core import setup
from setuptools import find_packages

setup(
    name='django-instapush',
    version='1.0.0',
    author='Amyth Arora',
    author_email='mail@amythsingh.com',
    packages=find_packages(),
    url='https://github.com/amyth/django-instapush',
    license='MIT License',
    description='Send push notifications to android and ios devices',
    long_description='Django Instapush can be used to send GCM and APNS '\
            'notifications to android and ios devices respectively. This '\
            'package supports both mysql and mongoengine models to store '\
            'device data',
    zip_safe=False,
)



from django.db import models
from mongoengine.queryset import QuerySet


class GCMMessageMixin(object):
    def send_message(self, message, **kwargs):
        if self:
            from instapush.libs.gcm import gcm_send_bulk_message
            data = kwargs.pop("extra", {})
            if message is not None:
                data["message"] = message

            ids = [device.registration_id for device in self]
            return gcm_send_bulk_message(registration_ids=ids,
                    data=data, **kwargs)

class APNSMessageMixin(object):
    def send_message(self, message, **kwargs):
        from instapush.libs.apns import apns_send_bulk_message
        ids = [device.registration_id for device in self]

        return apns_send_bulk_message(registration_ids=ids,
                alert=message, **kwargs)


class GCMQuerySet(GCMMessageMixin, models.query.QuerySet):
    """
    Implements additional methods to be used by this queryset.
    """
    pass


class APNSQuerySet(APNSMessageMixin, models.query.QuerySet):
    """
    Implements additional methods to be used by this queryset.
    """
    pass


class GCMMongoQuerySet(GCMMessageMixin, QuerySet):
    """
    Defines additional methods to be used by GCM mongo query set
    """
    pass


class APNSMongoQuerySet(APNSMessageMixin, QuerySet):
    """
    Defines additional methods to be used by APNS mongo query set
    """
    pass




"""
A stand-alone library to send GCM push notifications
"""

import json

## import urllib methods
try:
    from urllib.request import Request, urlopen
    from urllib.parse import urlencode
except ImportError:
    from urllib2 import Request, urlopen
    from urllib import urlencode

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured

from ..exceptions import GCMPushError
from ..settings import INSTAPUSH_SETTINGS as settings


class GCMMessenger(object):

    def __init__(self, registration_id, data, encoding='utf-8', **kwargs):

        self._registration_id = registration_id
        self._data = data
        self._kwargs = kwargs
        self.encoding = encoding

        ## Check for required settings
        self._prepare_settings()

    def _chunks(self):
        for item in range(0, len(self.registration_id), self.max_recipients):
            yield self.registration_id[item:item+self.max_recipients]

    def _prepare_settings(self):

        cons = ['API_KEY',]
        gcm_settings = settings.get('GCM_SETTINGS')

        for x in cons:
            item = gcm_settings.get(x)
            if not item:
                raise ImproperlyConfigured("Please add %s to your "\
                        "GCM_SETTINGS to send notifications through gcm" % x)

        for x in gcm_settings:
            setattr(self, x.lower(), gcm_settings[x])

    def send_plain(self):
        """
        Sends a text/plain GCM message
        """

        values = {"registration_id": self._registration_id}

        for key, val in self._data.items():
            values["data.%s" % (key)] = val.encode(self.encoding)

        for key, val in self._kwargs.items():
            if val and isinstance(val, bool):
                val = 1
                values[key] = val

        data = urlencode(sorted(values.items())).encode(self.encoding)
        result = self._send(data, "application/x-www-form-urlencoded;charset=UTF-8")

        if result.startswith("Error="):
            if result in ("Error=NotRegistered", "Error=InvalidRegistration"):
                ## TODO: deactivate the unregistered device
                return result

            raise GCMPushError(result)

        return result

    def send_json(self, ids=None):
        """
        Sends a json GCM message
        """

        items = ids or self._registration_id
        values = {"registration_ids": items}

        if self._data is not None:
            values["data"] = self._data

        for key, val in self._kwargs.items():
            if val:
                values[key] = val

        data = json.dumps(values, separators=(",", ":"), sort_keys=True).encode(
                self.encoding)

        result = json.loads(self._send(data, "application/json"))

        if ("failure" in result) and (result["failure"]):
            unregistered = []
            throw_error = False

            for index, error in enumerate(result.get("results", [])):
                error = error.get("error", "")
                if error in ("NotRegistered", "InvalidRegistration"):
                    unregistered.append(items[index])
                elif error != "":
                    throw_error = True

            self.deactivate_unregistered_devices(unregistered)
            if throw_error:
                raise GCMPushError(result)

        return result

    def send_bulk(self):
        if len(self._registration_id) > self.max_recipients:
            result = []
            for chunk in self._chunks():
                result.append(self.send_json(ids=chunk))

            return result
        return self.send_json()

    def deactivate_unregistered_devices(self, rids):
        deactivate_callback = getattr(settings, 'DEACTIVATE_UNREG_CALLBACK')
        deactivate_callback(rids)


    def _send(self, data, content_type):
        """
        Sends a GCM message with the given content type
        """

        headers = {
            "Content-Type": content_type,
            "Authorization": "key=%s" % (self.api_key),
            "Content-Length": str(len(data))
        }

        request = Request(self.api_url, data, headers)
        return urlopen(request).read().decode(self.encoding)


def gcm_send_message(registration_id, data, encoding='utf-8', **kwargs):
    """
    Standalone method to send a single gcm notification
    """

    messenger = GCMMessenger(registration_id, data, encoding=encoding, **kwargs)
    return messenger.send_plain()


def gcm_send_bulk_message(registration_ids, data, encoding='utf-8', **kwargs):
    """
    Standalone method to send bulk gcm notifications
    """

    messenger = GCMMessenger(registration_ids, data, encoding=encoding, **kwargs)
    return messenger.send_bulk()



import mongoengine

from django.conf import settings
from django.utils import timezone

from .querysets import APNSMongoQuerySet, GCMMongoQuerySet
from ..settings import INSTAPUSH_SETTINGS as instapush_settings
from ..utils import get_model


class BaseOwner(mongoengine.Document):
    pass


class BaseDevice(mongoengine.Document):
    """
    A basic device document. This class defines the generic
    fields to be used by all device models/documents. All
    other device models should inherit this base device doc.
    """

    name = mongoengine.StringField()
    active = mongoengine.BooleanField(default=True)
    owner = mongoengine.ReferenceField(get_model(instapush_settings.get(
        'DEVICE_OWNER_MODEL')), required=False)
    date_created = mongoengine.DateTimeField(default=timezone.now())

    meta = {'collection': 'devices', 'allow_inheritance': True}


class GCMDevice(BaseDevice):
    """
    This document represents a device that is uses google cloud
    messaging service aka GCM. In simple words this document
    represents an android device.
    """

    device_id = mongoengine.StringField()
    registration_id = mongoengine.StringField()

    meta = {
        'queryset_class': GCMMongoQuerySet,
        'indexes': [{'fields': ['device_id'], 'unique': True, 'sparse': True}]
    }

    def send_message(self, data, **kwargs):
        from ..libs.gcm import gcm_send_message

        extra_data = kwargs.pop("extra", {})
        data.update(extra_data)

        return gcm_send_message(registration_id=self.registration_id,
                data=data, **kwargs)


class APNSDevice(BaseDevice):
    """
    This document represents an iOS device that uses APNS
    to send push notifications.
    """

    device_id = mongoengine.StringField()
    registration_id = mongoengine.StringField()

    meta = {
        'queryset_class': APNSMongoQuerySet,
        'indexes': [{'fields': ['device_id'], 'unique': True, 'sparse': True}]
    }

    def send_message(self, message, **kwargs):
        from ..libs.apns import apns_send_message
        return apns_send_message(registration_id=self.registration_id,
                alert=message, **kwargs)
